#include <GlewOpenGL.h>
#include "Viewer.h"
#include <ApiGraphics/SceneOBJ.h>
#include <ApiGraphics/ProgGLSL.h>
#include <ApiGraphics/Renderable.h>
#include <ApiGraphics/Message.h>
#include <QKeyEvent>
#include <math.h>

using namespace apig;

//using namespace std;
using qglviewer::Vec;
using qglviewer::Camera;
int textureUnit = 0;

const QList<QColor> Viewer::backgrounds = QList<QColor>() << QColor(45,45,45) << QColor(255,255,255) << QColor(0,0,0);

/**************************************************************************************************/

namespace apig {
    /** Render a flat square: the ground, for shadow map projection. */
    class RenderableGround : public Renderable {
    public:
        RenderableGround(float r, float z) : r(r), z(z) {}

        virtual void render() const;

        //virtual Vec3  center() const;
        virtual float radius() const;

    private:
        float r;
        float z;
    };

    /*********************************************************************************************/
    // RenderableGround implementation :

    float RenderableGround::radius() const { return sqrt(z*z + 2*r*r); }

    void RenderableGround::render() const {
        glBegin(GL_QUADS);
        glNormal3f(0,1,0);
        glVertex3f( r, z, r);
        glVertex3f( r, z, -r);
        glVertex3f(-r, z, -r);
        glVertex3f(-r, z, r);
        glEnd();
    }


}

/**************************************************************************************************/

Viewer::Viewer(QString fileName, QWidget *parent)
    : QGLViewer(QGLFormat(QGL::DoubleBuffer | QGL::DepthBuffer | QGL::Rgba | QGL::AlphaChannel | QGL::StencilBuffer), parent),
    object(NULL),
    wireframe(false),
    bgrIndex(0), shadowMapWidth(2048), shadowMapHeight(2048), debugBuffer(0), shadowProg(NULL)
{
    // Load the object from file:
    SceneOBJ* obj = new SceneOBJ(fileName);
    if (!obj->loaded()) {
        Message::error("cannot load specified object");
        exit(1);
    }
    object = obj;

    // Initialize the gound
    apig::RenderableGround* grnd = new RenderableGround(5*object->radius(), -0.48*object->radius());
    ground = grnd;

    // Light source initial position
    lightFrame = new qglviewer::ManipulatedFrame();
    lightFrame->setPosition(0.4f, 4.0f, 3.0f);
    lightCamera = new qglviewer::Camera();
    objectFrame = new qglviewer::ManipulatedFrame();

    setSnapshotFormat("PNG");           // format par defaut pour les images sauvegardees
    setAnimationPeriod(0);              // pas de limitation pour les FPS
    setStateFileName(QString::null);    // pas de sauvegarde d'etat dans un fichier xml
}

Viewer::~Viewer() {
    if (object != NULL) delete object;
    if (ground != NULL) delete ground;
    if (lightFrame != NULL) delete lightFrame;
    if (objectFrame != NULL) delete objectFrame;
    if (lightCamera != NULL) delete lightCamera;

    // delete fbo
    glDeleteFramebuffersEXT(1, &fboId);
    // delete depth texture
    glDeleteTextures(1, &depthbuffer);
    // delete colorBuffer
    glDeleteTextures(1, &colorbuffer);

    if (shadowProg) delete shadowProg;
}

/**************************************************************************************************/

void Viewer::keyPressEvent(QKeyEvent *e) {
    switch (e->key()) {
    case Qt::Key_W:
        wireframe = !wireframe;
        updateGL();
        break;

    case Qt::Key_Backspace :
        setBackgroundColor(backgrounds[bgrIndex = ((bgrIndex + 1) % backgrounds.size())]);
        updateGL();
        break;

    case Qt::Key_D:
        debugBuffer = !debugBuffer;
        updateGL();
        break;

    case Qt::Key_L:
        setManipulatedFrame( lightFrame );
        break;
    case Qt::Key_O:
        setManipulatedFrame( objectFrame );
        break;

    default:
        QGLViewer::keyPressEvent(e);
    }
}

void Viewer::generateShadowFBO() {

    // create a framebuffer object and two associated buffers
    glGenFramebuffersEXT(1, &fboId);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fboId);

    // setup color buffer:
    glGenTextures(1, &colorbuffer);
    glBindTexture(GL_TEXTURE_2D, colorbuffer);
    // Framebuffer texture initialization
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, shadowMapWidth, shadowMapHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, colorbuffer, 0);

    // create the depth texture we'll use for the shadowmap
    glGenTextures(1, &depthbuffer);
    glBindTexture(GL_TEXTURE_2D, depthbuffer);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    // Remove artifact on the edges of the shadowmap
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
    glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, shadowMapWidth, shadowMapHeight, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, depthbuffer, 0);

    // check FBO status
    // fbo-creation error-checking
    GLenum status =  glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
    if (status != GL_FRAMEBUFFER_COMPLETE_EXT) {
        std::cout << "GL_FRAMEBUFFER_COMPLETE_EXT failed, CANNOT use FBO" << std::endl;
        // error
    }
    // switch back to std. display.
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);  // std. rendering
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, colorbuffer);

    // And while we're at it, load the shader program:
    ProgGLSL *newProg = ProgGLSL::loadDir("shaders/blinnPhong/"); // blinnPhong
    if (newProg != NULL) {
        if (shadowProg != NULL) delete shadowProg;
        shadowProg = newProg;
        // depth buffer will be in texture unit 0.
        shadowProg->setUniform("ShadowMap", &textureUnit);
    }
}

/**************************************************************************************************/

void Viewer::init() {

    OpenGL::init();
    object->init();
    setSceneCenter(Vec(object->center()));
    setSceneRadius(object->radius());
    showEntireScene();
    setSceneRadius(ground->radius());

    setBackgroundColor(backgrounds[bgrIndex]);
    generateShadowFBO();
}

/**************************************************************************************************/

void inverseGLmatrix(double direct[16], float inverse[16])
{
    float m00 = direct[0];
    float m01 = direct[1];
    float m02 = direct[2];
    float m03 = direct[3];
    float m10 = direct[4];
    float m11 = direct[5];
    float m12 = direct[6];
    float m13 = direct[7];
    float m20 = direct[8];
    float m21 = direct[9];
    float m22 = direct[10];
    float m23 = direct[11];
    float m30 = direct[12];
    float m31 = direct[13];
    float m32 = direct[14];
    float m33 = direct[15];

    double determinant =
    m03 * m12 * m21 * m30-m02 * m13 * m21 * m30-m03 * m11 * m22 * m30+m01 * m13    * m22 * m30+
    m02 * m11 * m23 * m30-m01 * m12 * m23 * m30-m03 * m12 * m20 * m31+m02 * m13    * m20 * m31+
    m03 * m10 * m22 * m31-m00 * m13 * m22 * m31-m02 * m10 * m23 * m31+m00 * m12    * m23 * m31+
    m03 * m11 * m20 * m32-m01 * m13 * m20 * m32-m03 * m10 * m21 * m32+m00 * m13    * m21 * m32+
    m01 * m10 * m23 * m32-m00 * m11 * m23 * m32-m02 * m11 * m20 * m33+m01 * m12    * m20 * m33+
    m02 * m10 * m21 * m33-m00 * m12 * m21 * m33-m01 * m10 * m22 * m33+m00 * m11    * m22 * m33;

    inverse[0] = (m12*m23*m31 - m13*m22*m31 + m13*m21*m32 - m11*m23*m32 - m12*m21*m33 + m11*m22*m33)/determinant;
    inverse[1] = (m03*m22*m31 - m02*m23*m31 - m03*m21*m32 + m01*m23*m32 + m02*m21*m33 - m01*m22*m33)/determinant;
    inverse[2] = (m02*m13*m31 - m03*m12*m31 + m03*m11*m32 - m01*m13*m32 - m02*m11*m33 + m01*m12*m33)/determinant;
    inverse[3] = (m03*m12*m21 - m02*m13*m21 - m03*m11*m22 + m01*m13*m22 + m02*m11*m23 - m01*m12*m23)/determinant;
    inverse[4] = (m13*m22*m30 - m12*m23*m30 - m13*m20*m32 + m10*m23*m32 + m12*m20*m33 - m10*m22*m33)/determinant;
    inverse[5] = (m02*m23*m30 - m03*m22*m30 + m03*m20*m32 - m00*m23*m32 - m02*m20*m33 + m00*m22*m33)/determinant;
    inverse[6] = (m03*m12*m30 - m02*m13*m30 - m03*m10*m32 + m00*m13*m32 + m02*m10*m33 - m00*m12*m33)/determinant;
    inverse[7] = (m02*m13*m20 - m03*m12*m20 + m03*m10*m22 - m00*m13*m22 - m02*m10*m23 + m00*m12*m23)/determinant;
    inverse[8] = (m11*m23*m30 - m13*m21*m30 + m13*m20*m31 - m10*m23*m31 - m11*m20*m33 + m10*m21*m33)/determinant;
    inverse[9] = (m03*m21*m30 - m01*m23*m30 - m03*m20*m31 + m00*m23*m31 + m01*m20*m33 - m00*m21*m33)/determinant;
    inverse[10] = (m01*m13*m30 - m03*m11*m30 + m03*m10*m31 - m00*m13*m31 - m01*m10*m33 + m00*m11*m33)/determinant;
    inverse[11] = (m03*m11*m20 - m01*m13*m20 - m03*m10*m21 + m00*m13*m21 + m01*m10*m23 - m00*m11*m23)/determinant;
    inverse[12] = (m12*m21*m30 - m11*m22*m30 - m12*m20*m31 + m10*m22*m31 + m11*m20*m32 - m10*m21*m32)/determinant;
    inverse[13] = (m01*m22*m30 - m02*m21*m30 + m02*m20*m31 - m00*m22*m31 - m01*m20*m32 + m00*m21*m32)/determinant;
    inverse[14] = (m02*m11*m30 - m01*m12*m30 - m02*m10*m31 + m00*m12*m31 + m01*m10*m32 - m00*m11*m32)/determinant;
    inverse[15] = (m01*m12*m20 - m02*m11*m20 + m02*m10*m21 - m00*m12*m21 - m01*m10*m22 + m00*m11*m22)/determinant;
}


void Viewer::draw() {

    glEnable(GL_TEXTURE_2D);
    GLdouble objectMatrix[16];
    objectFrame->getMatrix(objectMatrix);
    // store previous viewport
    glPushAttrib(GL_VIEWPORT_BIT);
        glDisable(GL_CULL_FACE);
        //First step: Render from the light POV to a FBO, story depth values only
        // bind fbo
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fboId); // rendering offscreen
        // Clear previous frame values
        glViewport(0, 0, shadowMapWidth, shadowMapHeight);
        glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
        //Using the fixed pipeline to render to the depthbuffer
        shadowProg->deactivate();
        // In the case we render the shadowmap to a higher resolution, the viewport must be modified accordingly.
        // Prepare the shadow map.
        // The light camera needs size hints in order to optimize its fieldOfView
        lightCamera->setSceneRadius(sceneRadius());
        lightCamera->setSceneCenter(sceneCenter());
        // Place the light camera.
        lightCamera->setPosition(lightFrame->position());
        lightCamera->lookAt(sceneCenter());
        lightCamera->setFOVToFitScene();
        //Disable color rendering, we only want to write to the Z-Buffer
        glDisable(GL_TEXTURE_2D);
        if (!debugBuffer) glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
        else glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
        glDepthMask(GL_TRUE);
        // set up projection as Camera proj:
        lightCamera->loadProjectionMatrix();
        lightCamera->loadModelViewMatrix();
        glMatrixMode(GL_MODELVIEW);
        if (debugBuffer) {
             glColor4f(1.0, 1.0, 1.0, 1.0);
         }
        // render objects
        glPushMatrix();
            glMultMatrixd(objectMatrix);
            object->render();
        glPopMatrix();
        ground->render();
        // At this point, we have a texture with the model
        if (debugBuffer) {
            glBindTexture(GL_TEXTURE_2D, colorbuffer);
            glGenerateMipmapEXT(GL_TEXTURE_2D);
        }
        glFlush();
    glPopAttrib();
    // Now rendering from the camera POV, using the FBO to generate shadows
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
    //Enabling color write (previously disabled for light POV z-buffer rendering)
    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, depthbuffer);

    // We reload these because the camera might have changed: no need to push/pop
    // GL_PROJECTION matrix
    camera()->loadProjectionMatrix();
    // GL_MODELVIEW matrix
    camera()->loadModelViewMatrix();
    shadowProg->activate();
    shadowProg->setUniformVec3("lightSource", lightFrame->position()[0], lightFrame->position()[1], lightFrame->position()[2]);
    shadowProg->setUniformVec4("cameraPosition", camera()->position()[0], camera()->position()[1], camera()->position()[2], camera()->position()[3]);
    GLdouble cameraWorld[16];
    camera()->getModelViewMatrix(cameraWorld);
    float cameraInverse[16];
    inverseGLmatrix(cameraWorld, cameraInverse);
    shadowProg->setUniformMat4("cameraInverse", cameraInverse);
    // Fin debug
    float cameraDirect[16];
    for (int i=0; i < 16; i++) cameraDirect[i] = cameraWorld[i];
    shadowProg->setUniformMat4("cameraDirect", cameraDirect);
    GLdouble textureMatrix[16];
    lightCamera->getModelViewProjectionMatrix(textureMatrix);
    glMatrixMode(GL_TEXTURE);
    glLoadMatrixd(textureMatrix);
    glMatrixMode(GL_MODELVIEW);
    // Draw the scene itself.
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glPolygonMode(GL_FRONT_AND_BACK, wireframe ? GL_LINE : GL_FILL);
    glColor3f(0.6, 0.7, 0.8);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
        glMultMatrixd(objectMatrix);
        object->render();
    glPopMatrix();
    glColor3f(0.9f, 0.9f, 0.9f);
    ground->render();

    if (debugBuffer) {
        // This piece of code draws the depth buffer onscreen
        glViewport(0, 0, width(), height());
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        gluOrtho2D(0, width(), 0, height());
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        shadowProg->deactivate();
        // show depth texture
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, depthbuffer);
        glColor3f(0.8f, 0.8f, 0.8f);
        glBegin(GL_QUADS);
            glTexCoord2f(0, 0); glVertex2f(width() - height()/2, 0);
            glTexCoord2f(1, 0); glVertex2f(width(),   0);
            glTexCoord2f(1, 1); glVertex2f(width(),   height()/2);
            glTexCoord2f(0, 1); glVertex2f(width() - height()/2, height()/2);
        glEnd();
        glEnable(GL_DEPTH_TEST);
    }
}

