#ifndef __COLOR_TYPES__
#define __COLOR_TYPES__

#include <GlewOpenGL.h>
#include <QColor>

namespace apig {

    template<class Color>
    class GenericColor {
        public:
            static Color create(float r, float g, float b, float a) { Color res; res.reset(r,g,b,a); return res; }
            static Color create(float c) { return create(c,c,c,c); }
            static Color BLACK() { return create(0); }
            static Color WHITE() { return create(1); }
            
            GenericColor(Color &color) : color(color) {}
            
            float norm2() const {
                float e = 0.0f;
                for (int c=0; c<Color::SIZE; c++) e += color[c] * color[c];
                return e;
                }
    
            float norm() const { return sqrt(norm2()); }
            float normRMS() const { return sqrt(norm2() / Color::SIZE); }
            
        private:
            Color &color;
        };
    
    class Float1 {
        public:
            Float1(float x = 0);
            Float1(const float *c);   // tableau de 1 valeur
            Float1(QRgb c);
        
            // operateurs de conversion :
            inline operator const float* () const {return c;}     // cast du Float1 en const float*
            inline operator       float* ()       {return c;}     // cast du Float1 en float*
            inline operator const float  () const {return r;}     // cast du Float1 en const float
            inline operator       float  ()       {return r;}     // cast du Float1 en float
            QRgb toQRgb() const;
            QString toQString() const { return QString("(%1)").arg(r); }
    
            //operateurs arithmetiques unaires :
            friend Float1 operator-(const Float1 &a);
        
            // operateurs arithmetiques binaires (membres a membres) :
            friend Float1 operator+(const Float1 &a, const Float1 &b);
            friend Float1 operator-(const Float1 &a, const Float1 &b);
            friend Float1 operator*(const Float1 &a, const Float1 &b);
            friend Float1 operator/(const Float1 &a, const Float1 &b);
                
            // operations avec un scalaire :
            friend Float1 operator*(float s, const Float1 &a);
            friend Float1 operator*(const Float1 &a, float s);
            friend Float1 operator/(float s, const Float1 &a);
            friend Float1 operator/(const Float1 &a, float s);
            
            // operateurs de modification :
            Float1 operator+=(const Float1 &a);
            Float1 operator-=(const Float1 &a);
            Float1 operator*=(const Float1 &a);
            Float1 operator/=(const Float1 &a);
            Float1 operator*=(float s);
            Float1 operator/=(float s);
    
            void reset(float r=0, float g=0, float b=0, float a=0);
            void setGL() const;
            
        public:
            union {
                struct { float r; };
                struct { float x; };
                float c[1];
                };
            static const int SIZE = 1;
            static const GLint  TEX_FORMAT = 1;  // = GL_LUMINANCE
            static const GLenum DATA_FORMAT = GL_LUMINANCE;
            static const GLenum DATA_TYPE = GL_FLOAT;
        };
    
    class Float2 {
        public:
            Float2(float x = 0, float y = 0);
            Float2(const float *c);   // tableau de 2 valeurs
            Float2(QRgb c);
        
            // operateurs de conversion :
            inline operator const float* () const {return c;}     // cast du Float2 en const float*
            inline operator       float* ()       {return c;}     // cast du Float2 en float*
            QRgb toQRgb() const;
            QString toQString() const { return QString("(%1, %2)").arg(r).arg(g); }
    
            //operateurs arithmetiques unaires :
            friend Float2 operator-(const Float2 &a);
        
            // operateurs arithmetiques binaires (membres a membres) :
            friend Float2 operator+(const Float2 &a, const Float2 &b);
            friend Float2 operator-(const Float2 &a, const Float2 &b);
            friend Float2 operator*(const Float2 &a, const Float2 &b);
            friend Float2 operator/(const Float2 &a, const Float2 &b);
                
            // operations avec un scalaire :
            friend Float2 operator*(float s, const Float2 &a);
            friend Float2 operator*(const Float2 &a, float s);
            friend Float2 operator/(float s, const Float2 &a);
            friend Float2 operator/(const Float2 &a, float s);
            
            // operateurs de modification :
            Float2 operator+=(const Float2 &a);
            Float2 operator-=(const Float2 &a);
            Float2 operator*=(const Float2 &a);
            Float2 operator/=(const Float2 &a);
            Float2 operator*=(float s);
            Float2 operator/=(float s);
    
            void reset(float r=0, float g=0, float b=0, float a=0);
            void setGL() const;
            
        public:
            union {
                struct { float r, g; };
                struct { float x, y; };
                float c[2];
                };
            static const int SIZE = 2;
            static const GLint  TEX_FORMAT = 2;  // = GL_LUMINANCE_ALPHA
            static const GLenum DATA_FORMAT = GL_LUMINANCE_ALPHA;
            static const GLenum DATA_TYPE = GL_FLOAT;
        };
    
    class Float3 {
        public:
            Float3(float x = 0, float y = 0, float z = 0);
            Float3(const float *c);   // tableau de 3 valeurs
            Float3(QRgb c);
        
            // operateurs de conversion :
            inline operator const float* () const {return c;}     // cast du Float3 en const float*
            inline operator       float* ()       {return c;}     // cast du Float3 en float*
            QRgb toQRgb() const;
            QString toQString() const { return QString("(%1, %2, %3)").arg(r).arg(g).arg(b); }
    
            //operateurs arithmetiques unaires :
            friend Float3 operator-(const Float3 &a);
        
            // operateurs arithmetiques binaires (membres a membres) :
            friend Float3 operator+(const Float3 &a, const Float3 &b);
            friend Float3 operator-(const Float3 &a, const Float3 &b);
            friend Float3 operator*(const Float3 &a, const Float3 &b);
            friend Float3 operator/(const Float3 &a, const Float3 &b);
                
            // operations avec un scalaire :
            friend Float3 operator*(float s, const Float3 &a);
            friend Float3 operator*(const Float3 &a, float s);
            friend Float3 operator/(float s, const Float3 &a);
            friend Float3 operator/(const Float3 &a, float s);
            
            // operateurs de modification :
            Float3 operator+=(const Float3 &a);
            Float3 operator-=(const Float3 &a);
            Float3 operator*=(const Float3 &a);
            Float3 operator/=(const Float3 &a);
            Float3 operator*=(float s);
            Float3 operator/=(float s);
    
            void reset(float r=0, float g=0, float b=0, float a=0);
            void setGL() const;
            
        public:
            union {
                struct { float r, g, b; };
                struct { float x, y, z; };
                float c[3];
                };
            static const int SIZE = 3;
            static const GLint  TEX_FORMAT = 3;  // = GL_RGB
            static const GLenum DATA_FORMAT = GL_RGB;
            static const GLenum DATA_TYPE = GL_FLOAT;
        };
    
    class Float4 {
        public:
            Float4(float x = 0, float y = 0, float z = 0, float w = 1);
            Float4(const float *c);   // tableau de 4 valeurs
            Float4(const float *xy, const float *zw);   // 2 tableaux de 2 valeurs
            Float4(QRgb c);
            
            // operateurs de conversion :
            inline operator const float* () const {return c;}     // cast du Float4 en const float*
            inline operator       float* ()       {return c;}     // cast du Float4 en float*
            QRgb toQRgb() const;
            QString toQString() const { return QString("(%1, %2, %3, %4)").arg(r).arg(g).arg(b).arg(a); }
    
            //operateurs arithmetiques unaires :
            friend Float4 operator-(const Float4 &a);
        
            // operateurs arithmetiques binaires (membres a membres) :
            friend Float4 operator+(const Float4 &a, const Float4 &b);
            friend Float4 operator-(const Float4 &a, const Float4 &b);
            friend Float4 operator*(const Float4 &a, const Float4 &b);
            friend Float4 operator/(const Float4 &a, const Float4 &b);
                
            // operations avec un scalaire :
            friend Float4 operator*(float s, const Float4 &a);
            friend Float4 operator*(const Float4 &a, float s);
            friend Float4 operator/(float s, const Float4 &a);
            friend Float4 operator/(const Float4 &a, float s);
            
            // operateurs de modification :
            Float4 operator+=(const Float4 &a);
            Float4 operator-=(const Float4 &a);
            Float4 operator*=(const Float4 &a);
            Float4 operator/=(const Float4 &a);
            Float4 operator*=(float s);
            Float4 operator/=(float s);
    
            void reset(float r=0, float g=0, float b=0, float a=0);
            void setAlpha(float a);
            void setGL() const;
            
        public:
            union {
                struct { float r, g, b, a; };
                struct { float x, y, z, w; };
                float c[4];
                };
            static const int SIZE = 4;
            static const GLint  TEX_FORMAT = 4;  // = GL_RGBA
            static const GLenum DATA_FORMAT = GL_RGBA;
            static const GLenum DATA_TYPE = GL_FLOAT;
        };
    
    class UByte3;
    
    class UInt3 {
        public:
            UInt3() : x(0), y(0), z(0) {}
            UInt3(GLuint x, GLuint y, GLuint z);
            UInt3(const GLuint *c);   // tableau de 3 valeurs
            UInt3(QRgb c);
        
            // operateurs de conversion :
            inline operator const GLuint* () const {return c;}     // cast du UInt3 en const GLuint*
            inline operator       GLuint* ()       {return c;}     // cast du UInt3 en GLuint*
            QRgb toQRgb() const;
            QString toQString() const { return QString("(%1, %2, %3)").arg(r).arg(g).arg(b); }
    
            // operateurs arithmetiques binaires (membres a membres) :
            friend UInt3 operator+(const UInt3 &a, const UInt3 &b);
            friend UInt3 operator-(const UInt3 &a, const UInt3 &b);
            friend UInt3 operator*(const UInt3 &a, const UInt3 &b);
            friend UInt3 operator/(const UInt3 &a, const UInt3 &b);
                
            // operations avec un scalaire :
            friend UInt3 operator*(float s, const UInt3 &a);
            friend UInt3 operator*(const UInt3 &a, float s);
            friend UInt3 operator/(float s, const UInt3 &a);
            friend UInt3 operator/(const UInt3 &a, float s);
            
            // operateurs de modification :
            UInt3 operator+=(const UInt3 &a);
            UInt3 operator-=(const UInt3 &a);
            UInt3 operator*=(const UInt3 &a);
            UInt3 operator/=(const UInt3 &a);
            UInt3 operator*=(float s);
            UInt3 operator/=(float s);
    
            void reset(float r=0, float g=0, float b=0, float a=0);
            void setGL() const;
            
            UByte3 toUByte3() const;
            
        public:
            union {
                struct { GLuint r, g, b; };
                struct { GLuint x, y, z; };
                GLuint c[3];
                };
            static const int SIZE = 3;
            static const GLint  TEX_FORMAT = 3;  // = GL_RGB
            static const GLenum DATA_FORMAT = GL_RGB;
            static const GLenum DATA_TYPE = GL_UNSIGNED_INT;
        };
    
    class UByte4;
    
    class UInt4 {
        public:
            UInt4() : x(0), y(0), z(0), w(255) {}
            UInt4(GLuint x, GLuint y, GLuint z, GLuint w);
            UInt4(const GLuint *c);   // tableau de 4 valeurs
            UInt4(QRgb c);
        
            // operateurs de conversion :
            inline operator const GLuint* () const {return c;}     // cast du UInt4 en const GLuint*
            inline operator       GLuint* ()       {return c;}     // cast du UInt4 en GLuint*
            QRgb toQRgb() const;
            QString toQString() const { return QString("(%1, %2, %3, %4)").arg(r).arg(g).arg(b).arg(a); }
    
            // operateurs arithmetiques binaires (membres a membres) :
            friend UInt4 operator+(const UInt4 &a, const UInt4 &b);
            friend UInt4 operator-(const UInt4 &a, const UInt4 &b);
            friend UInt4 operator*(const UInt4 &a, const UInt4 &b);
            friend UInt4 operator/(const UInt4 &a, const UInt4 &b);
                
            // operations avec un scalaire :
            friend UInt4 operator*(float s, const UInt4 &a);
            friend UInt4 operator*(const UInt4 &a, float s);
            friend UInt4 operator/(float s, const UInt4 &a);
            friend UInt4 operator/(const UInt4 &a, float s);
            
            // operateurs de modification :
            UInt4 operator+=(const UInt4 &a);
            UInt4 operator-=(const UInt4 &a);
            UInt4 operator*=(const UInt4 &a);
            UInt4 operator/=(const UInt4 &a);
            UInt4 operator*=(float s);
            UInt4 operator/=(float s);
    
            void reset(float r=0, float g=0, float b=0, float a=0);
            void setAlpha(float a);
            void setGL() const;
            
            UByte4 toUByte4() const;
            
        public:
            union {
                struct { GLuint r, g, b, a; };
                struct { GLuint x, y, z, w; };
                GLuint c[4];
                };
            static const int SIZE = 4;
            static const GLint  TEX_FORMAT = 4;  // = GL_RGBA
            static const GLenum DATA_FORMAT = GL_RGBA;
            static const GLenum DATA_TYPE = GL_UNSIGNED_INT;
        };
    
    class UByte3 {
        public:
            UByte3(GLubyte x = 0, GLubyte y = 0, GLubyte z = 0);
            UByte3(const GLubyte *c);   // tableau de 3 valeurs
            UByte3(QRgb c);
        
            // operateurs de conversion :
            inline operator const GLubyte* () const {return c;}     // cast du UByte3 en const GLubyte*
            inline operator       GLubyte* ()       {return c;}     // cast du UByte3 en GLubyte*
            QRgb toQRgb() const;
            QString toQString() const { return QString("(%1, %2, %3)").arg(r).arg(g).arg(b); }
    
            // operateurs arithmetiques binaires (membres a membres) :
            friend UByte3 operator+(const UByte3 &a, const UByte3 &b);
            friend UByte3 operator-(const UByte3 &a, const UByte3 &b);
            friend UByte3 operator*(const UByte3 &a, const UByte3 &b);
            friend UByte3 operator/(const UByte3 &a, const UByte3 &b);
                
            // operations avec un scalaire :
            friend UByte3 operator*(float s, const UByte3 &a);
            friend UByte3 operator*(const UByte3 &a, float s);
            friend UByte3 operator/(float s, const UByte3 &a);
            friend UByte3 operator/(const UByte3 &a, float s);
            
            // operateurs de modification :
            UByte3 operator+=(const UByte3 &a);
            UByte3 operator-=(const UByte3 &a);
            UByte3 operator*=(const UByte3 &a);
            UByte3 operator/=(const UByte3 &a);
            UByte3 operator*=(float s);
            UByte3 operator/=(float s);
    
            void reset(float r=0, float g=0, float b=0, float a=0);
            void setGL() const;
            
            UInt3 toUInt3() const;
            
        public:
            union {
                struct { GLubyte r, g, b; };
                struct { GLubyte x, y, z; };
                GLubyte c[3];
                };
            static const int SIZE = 3;
            static const GLint  TEX_FORMAT = 3;  // = GL_RGB
            static const GLenum DATA_FORMAT = GL_RGB;
            static const GLenum DATA_TYPE = GL_UNSIGNED_BYTE;
        };
    
    class UByte4 {
        public:
            UByte4(GLubyte x = 0, GLubyte y = 0, GLubyte z = 0, GLubyte w = 255);
            UByte4(const GLubyte *c);   // tableau de 4 valeurs
            UByte4(QRgb c);
        
            // operateurs de conversion :
            inline operator const GLubyte* () const {return c;}     // cast du UByte4 en const GLubyte*
            inline operator       GLubyte* ()       {return c;}     // cast du UByte4 en GLubyte*
            QRgb toQRgb() const;
            QString toQString() const { return QString("(%1, %2, %3, %4)").arg(r).arg(g).arg(b).arg(a); }
    
            // operateurs arithmetiques binaires (membres a membres) :
            friend UByte4 operator+(const UByte4 &a, const UByte4 &b);
            friend UByte4 operator-(const UByte4 &a, const UByte4 &b);
            friend UByte4 operator*(const UByte4 &a, const UByte4 &b);
            friend UByte4 operator/(const UByte4 &a, const UByte4 &b);
                
            // operations avec un scalaire :
            friend UByte4 operator*(float s, const UByte4 &a);
            friend UByte4 operator*(const UByte4 &a, float s);
            friend UByte4 operator/(float s, const UByte4 &a);
            friend UByte4 operator/(const UByte4 &a, float s);
            
            // operateurs de modification :
            UByte4 operator+=(const UByte4 &a);
            UByte4 operator-=(const UByte4 &a);
            UByte4 operator*=(const UByte4 &a);
            UByte4 operator/=(const UByte4 &a);
            UByte4 operator*=(float s);
            UByte4 operator/=(float s);
    
            void reset(float r=0, float g=0, float b=0, float a=0);
            void setAlpha(float a);
            void setGL() const;
            
            UInt4 toUInt4() const;
            
        public:
            union {
                struct { GLubyte r, g, b, a; };
                struct { GLubyte x, y, z, w; };
                GLubyte c[4];
                };
            static const int SIZE = 4;
            static const GLint  TEX_FORMAT = 4;  // = GL_RGBA
            static const GLenum DATA_FORMAT = GL_RGBA;
            static const GLenum DATA_TYPE = GL_UNSIGNED_BYTE;
        };
    }

#endif

