#ifndef __IMAGE1D__
#define __IMAGE1D__

#include <GlewOpenGL.h>
#include "TextureData.h"
#include "ColorTypes.h"
#include "Texture.h"
#include <QImage>

namespace apig {

    /** This class is an abstract representation for ones-dimensional images. */
    class AbstractImage1D : public TextureData {
        public:
            AbstractImage1D(int w=0) : w(w) {}
            virtual ~AbstractImage1D() {}
    
            virtual bool loaded() const = 0;
            virtual void destroy() = 0;
    
            virtual QImage toQImage() const = 0;
            virtual void initialize(QImage image) = 0;       // initialise l'image  partir de <image>
    
            void save(QString fileName) const;              // sauvegarde l'image dans un format donne par l'extension du nom de fichier
            
            int width() const { return w; }
            int height() const { return h; }
            bool contains(int i)   const { return (i >= 0) && (i < w); }
            bool contains(float x) const { return (x >= 0) && (x < w); }
            
            virtual GLint defaultTexFormat() const = 0;
            virtual GLenum textureMode() const { return GL_TEXTURE_1D; }
            virtual void loadToGPU(GLint texFormat) const { loadTexture1D(texFormat, GL_TEXTURE_1D); }
            virtual void loadTexture1D(GLint texFormat, GLenum target = GL_TEXTURE_1D) const = 0;
    
            enum WrapMode { CLAMP_TO_EDGE, CLAMP_TO_BORDER, REPEAT, MIRRORED_REPEAT };
            
        protected:
            int w, h;
        };
    
    /////////////////////////////////////////////////
    
    /** This class represents a one-dimensional image.
     * It is parameterized by the data type representing colors (number of channels and type used per channel).
     * Useful types for color representations are defined in \c ColorTypes.h : \c Float1, \c Float2, \c Float3, \c Float4, \c UByte3, \c UByte4, \c UInt3, \c UInt4.
     */
    template<class Color>
    class Image1D : public AbstractImage1D {
        public:
            Image1D(int w=0, Color* data=NULL, WrapMode wrap=CLAMP_TO_EDGE, Color border=Color());
            Image1D(QImage image);
            Image1D(QString fileName);
            virtual ~Image1D() {}
    
            virtual bool loaded() const { return data != NULL; }
            virtual void destroy();
            
            virtual QImage toQImage() const;
            virtual void initialize(QImage image);      // initialise l'image  partir de <image>
            Image1D<Color> clone() const;               // duplique l'image en une nouvelle image
            void copy(const Image1D<Color> &image);     // initialise l'image  partir de <image>  (precondition : width() >= image.width())
            
            virtual GLint defaultTexFormat() const { return Color::TEX_FORMAT; }
            virtual void loadTexture1D(GLint texFormat, GLenum target = GL_TEXTURE_1D) const;
    
            // acquisition d'une texture 1D :
            static Image1D<Color> readTexture(Texture *tex);
    
            // definition de l'echantillonnage aux bords de l'image :
            void setBorderColor(Color border);
            void setWrapMode(WrapMode wrapMode);
            void setupBorder(WrapMode wrapMode, Color border);
            
            // fonctions d'acces aux pixels :
            //-------------------------------
            inline       Color& texel(int i)       { return data[i]; }          // precondition : this->contains(i)
            inline const Color& texel(int i) const { return data[i]; }          // precondition : this->contains(i)
            inline       Color& operator()(int i)       { return texel(i); }
            inline const Color& operator()(int i) const { return texel(i); }
            Color sample(int i) const;                                  // si !this->contains(i), le resultat depend de <borderColor> et de <wrapMode>
            Color interp(float x) const;                                // interpolation lineaire : image mappee sur [0,w[
            Color interpUnit(float x) const { return interp(x * w); }   // interpolation lineaire : image mappee sur [0,1[
            Color operator()(float x) const { return interp(x); }
            inline const Color* mem() const { return data; }
            inline       Color* mem()       { return data; }
    
        private:
            Color *data;
            WrapMode wrapMode;
            Color borderColor;
        };
    
    typedef Image1D<UByte4> Image1DUByte4;
    typedef Image1D<Float1> Image1DFloat;
    typedef Image1D<Float2> Image1DFloat2;
    typedef Image1D<Float3> Image1DFloat3;
    typedef Image1D<Float4> Image1DFloat4;
    
    }

#include "Image1D_impl.h"

#endif

