//#include "ImageFunctions.h"
#include "Util.h"
#include "ViewportGL.h"
#include "Message.h"
#include <math.h>

namespace apig {

/*********************************************************************************************/
// Image1D :

template<class Color>
Image1D<Color>::Image1D(int w, Color* inputData, WrapMode wrap, Color border) : AbstractImage1D(w), data(inputData), wrapMode(wrap), borderColor(border) {
    if (w <= 0) {
        w = 0;
        data = NULL;
        }
    else if (inputData == NULL) {
        data = new Color[w];
        if (data == NULL) {
            Message::error("l'allocation memoire pour l'image a echoue");
            w = 0;
            }
        }
    }

template<class Color>
Image1D<Color>::Image1D(QImage image) : AbstractImage1D(0), data(NULL), wrapMode(CLAMP_TO_EDGE) {
    initialize(image);
    }

template<class Color>
Image1D<Color>::Image1D(QString fileName) : AbstractImage1D(0), data(NULL), wrapMode(CLAMP_TO_EDGE) {
    QImage image(fileName);
    if (image.isNull())
        Message::error(QString("probleme lors du chargement du fichier '%1'").arg(fileName));
    else
        initialize(image);
    }

template<class Color>
void Image1D<Color>::destroy() {
    if (data != NULL) delete[] data;
    data = NULL;
    w = 0;
    }

template<class Color>
void Image1D<Color>::initialize(QImage image) {
    destroy();
    if (image.isNull())
        Message::error("l'image fournie ne contient pas de donnees");
    else {
        w = image.width();
        data = new Color[w];
        if (data == NULL) {
            Message::error("l'allocation memoire pour l'image a echoue");
            w = 0;
            }
        else
            for (int i=0; i<w; i++)
                texel(i) = Color(image.pixel(i, 0));
        }
    }

template<class Color>
QImage Image1D<Color>::toQImage() const {
    QImage image(w, 1, QImage::Format_ARGB32);
    for (int i=0; i<w; i++)
        image.setPixel(i, 0, texel(i).toQRgb());
    return image;
    }

template<class Color>
void Image1D<Color>::copy(const Image1D<Color> &image) {
    if (!image.loaded())
        Message::error("l'image fournie ne contient pas de donnees");
    else if (w < image.width())
        Message::error("l'image source est plus grande que l'image destination");
    else {
        for (int c=0; c<w; c++) data[c] = image.data[c];
        wrapMode = image.wrapMode;
        borderColor = image.borderColor;
        }
    }

template<class Color>
Image1D<Color> Image1D<Color>::clone() const {
    if (data == NULL) return Image1D<Color>();
    Color *clonedData = new Color[w];
    if (clonedData == NULL) {
        Message::error("l'allocation memoire pour l'image a echoue");
        return Image1D<Color>();
        }
    for (int c=0; c<w; c++)
        clonedData[c] = data[c];
    return Image1D<Color>(w, clonedData, wrapMode, borderColor);
    }

template<class Color>
void Image1D<Color>::setBorderColor(Color c) {
    borderColor = c;
    }

template<class Color>
void Image1D<Color>::setWrapMode(WrapMode mode) {
    wrapMode = mode;
    }

template<class Color>
void Image1D<Color>::setupBorder(WrapMode mode, Color c) {
    wrapMode = mode;
    borderColor = c;
    }

template<class Color>
Color Image1D<Color>::sample(int i) const {
    switch (wrapMode) {
        case CLAMP_TO_BORDER :  return contains(i) ? texel(i) : borderColor;
        case REPEAT :           return texel(util::modulo(i, w));
        case MIRRORED_REPEAT :  return texel(util::mirror(i, w));
        case CLAMP_TO_EDGE :
        default :               return texel(util::clamp(i, 0, w-1));
        };
    }

template<class Color>
Color Image1D<Color>::interp(float x) const {
    int i0 = (int)(floorf(x - 0.5));
    float a = x - 0.5 - i0;
    return (1 - a) * sample(i0) + a * sample(i0 + 1);
    }

/*********************************************************************************************/
// fonctions de lecture/ecriture de buffer OpenGL :

template<class Color>
void Image1D<Color>::loadTexture1D(GLint texFormat, GLenum target) const {
    glTexImage1D(target, 0, texFormat, w, 0, Color::DATA_FORMAT, Color::DATA_TYPE, data);
    }

template<class Color>
Image1D<Color> Image1D<Color>::readTexture(Texture *tex) {
    int w = tex->getWidth();
    Image1D<Color> res(w, NULL);
    if (w < 0)
        Message::error("dimensions de la texture non definies");
    else {
        tex->bind();
        glGetTexImage(GL_TEXTURE_1D, 0, Color::DATA_FORMAT, Color::DATA_TYPE, res.data);
        }
    return res;
    }

}

