#ifndef __IMAGE2D__
#define __IMAGE2D__

#include <GlewOpenGL.h>
#include "TextureData.h"
#include "ColorTypes.h"
#include "Texture.h"
#include "Image1D.h"
#include <QImage>

namespace apig {

    /** This class is an abstract representation for two-dimensional images.
     * It uses the same system of coordinates as OpenGL :
     * - \c i denotes the abscissa from left to right in the interval [0,width()-1]
     * - \c j denotes the ordinate from bottom to top in the interval [0, height()-1]
     *
     * So the origin (0,0) is at the leftmost bottommost corner of the image.
     @attention this system of coordinates if different from the one used in \c QImage (origin at the leftmost topmost corner)
     */
    class AbstractImage2D : public TextureData {
        public:
            AbstractImage2D(int w=0, int h=0) : w(w), h(h) {}
            virtual ~AbstractImage2D() {}
    
            virtual bool loaded() const = 0;    // true ssi l'image contient des donnees
            virtual void destroy() = 0;         // libre la memoire eventuellement occupee par l'image
            
            virtual QImage toQImage() const = 0;
            virtual void initialize(QImage image) = 0;      // initialise l'image a partir de <image>
            
            void save(QString fileName) const;              // sauvegarde l'image dans un format donne par l'extension du nom de fichier
    
            int width()  const { return w; }
            int height() const { return h; }
            bool contains(int i, int j)     const { return (i >= 0) && (i < w) && (j >= 0) && (j < h); }
            bool contains(float x, float y) const { return (x >= 0) && (x < w) && (y >= 0) && (y < h); }
            
            virtual GLenum textureMode() const { return GL_TEXTURE_2D; }
            virtual GLint defaultTexFormat() const = 0;
            virtual void loadToGPU(GLint texFormat) const { loadTexture2D(texFormat, GL_TEXTURE_2D); }
            virtual void loadTexture2D(GLenum texFormat, GLenum target = GL_TEXTURE_2D) const = 0;
    
            enum WrapMode { CLAMP_TO_EDGE, CLAMP_TO_BORDER, REPEAT, MIRRORED_REPEAT };
            
        protected:
            int w, h;
        };
    
    /////////////////////////////////////////////////
    
    /** This class represents a two-dimensional image.
     * It is parameterized by the data type representing colors (number of channels and type used per channel).
     * Useful types for color representations are defined in \c ColorTypes.h : \c Float1, \c Float2, \c Float3, \c Float4, \c UByte3, \c UByte4, \c UInt3, \c UInt4.
     */
    template<class Color>
    class Image2D : public AbstractImage2D {
        public:
            Image2D(int w=0, int h=0, Color* data=NULL);
            Image2D(QImage image);
            Image2D(QString fileName);
            virtual ~Image2D() {}
    
            virtual bool loaded() const { return data != NULL; }
            virtual void destroy();
            void clear(Color c = Color());
            
            virtual QImage toQImage() const;
            virtual void initialize(QImage image);      // initialise l'image a partir de <image>
            Image2D<Color> clone() const;               // duplique l'image en une nouvelle image
            Image2D<Color> subImage(int i, int j, int w, int h) const;  // precondition : i >= 0 && i+w <= width() && j >= 0 && j+h <= height()
            Image2D<Color> supImage(int i, int j, int w, int h) const;  // precondition : i >= 0 && i+width() <= w && j >= 0 && j+height() <= h
            Image2D<Color> boundingPowerOfTwo() const;  // retourne l'image englobant l'image courante, ayant des dimensions puissances de 2
            
            Image1D<Color> line(int j) const { return Image1D<Color>(w, data + j*w); }     // retourne l'Image1D referenant la ligne <j>
            
            virtual GLint defaultTexFormat() const { return Color::TEX_FORMAT; }
            virtual void loadTexture2D(GLenum texFormat, GLenum target = GL_TEXTURE_2D) const;
    
            // lecture/ecriture du buffer de couleurs OpenGL :
            static Image2D<Color> getColorBuffer();
            static Image2D<Color> getDepthBuffer();         // ne marche correctement que si le format <Color> a une seule composante
            void readColorBuffer(int i=0, int j=0);         // precondition : loaded() == true et le buffer contient le viewport (i,j,w,h)
            void readDepthBuffer(int i=0, int j=0);         // de mme mais le format <Color> doit avoir une seule composante
            void drawColorBuffer(int i=0, int j=0) const;   // point (i,j) en unites de pixels, origine en bas a gauche
            static Color readPixelColor(int i, int j);
            
            // acquisition d'une texture 2D :
            static Image2D<Color> getTexture(const Texture *tex);
            void readTexture(const Texture *tex);           // precondition : l'image a la mme taille que la texture
            
            // chargment direct d'une texture :
            static Texture createTex2D(QString fileName, GLint internalFormat = GL_RGBA, GLenum interpMode = GL_NEAREST, GLenum wrapMode = GL_CLAMP);
    
            // definition de l'echantillonnage aux bords de l'image :
            void setBorderColor(Color border);
            void setWrapMode(WrapMode wrapMode);
            void setupBorder(WrapMode wrapMode, Color border);
            
            // fonctions d'acces aux pixels :
            //-------------------------------
            inline       Color& texel(int i, int j)       { return data[i + j*w]; }    // precondition : this->contains(i,j)
            inline const Color& texel(int i, int j) const { return data[i + j*w]; }    // precondition : this->contains(i,j)
            inline       Color& operator()(int i, int j)       { return texel(i,j); }
            inline const Color& operator()(int i, int j) const { return texel(i,j); }
            Color sample(int i, int j) const;               // si !this->contains(i,j), le resultat depend de <borderColor> et de <wrapMode>
            Color interp(float x, float y) const;           // interpolation bilineaire : image mappee sur [0,w[*[0,h[
            inline Color operator()(float x, float y) const { return interp(x,y); }
            inline const Color* mem() const { return data; }
            inline       Color* mem()       { return data; }
            
        private:
            Color *data;
            Color borderColor;
            WrapMode wrapMode;
        };
    
    typedef Image2D<UByte4> Image2DUByte4;
    typedef Image2D<Float1> Image2DFloat;
    typedef Image2D<Float2> Image2DFloat2;
    typedef Image2D<Float3> Image2DFloat3;
    typedef Image2D<Float4> Image2DFloat4;
    
    }

#include "Image2D_impl.h"

#endif

