#include "Util.h"
#include "Message.h"
#include <QDir>
#include <math.h>

namespace apig {

/*********************************************************************************************/
// Image3D :

template<class Color>
Image3D<Color>::Image3D(int w, int h, int d, Color* inputData) : AbstractImage3D(w, h, d), data(inputData), wrapMode(CLAMP_TO_EDGE) {
    if (w <= 0 || h <= 0 || d <= 0) {
        w = h = d = 0;
        data = NULL;
        }
    else if (inputData == NULL) {
        data = new Color[w * h * d];
        if (data == NULL) {
            Message::error("l'allocation memoire pour l'image a echoue");
            w = h = d = 0;
            }
        }
    }

template<class Color>
void Image3D<Color>::destroy() {
    if (data != NULL) delete[] data;
    data = NULL;
    w = h = d = 0;
    }

template<class Color>
Image3D<Color> Image3D<Color>::clone() const {
    if (data == NULL) return Image3D<Color>();
    Color *clonedData = new Color[w * h * d];
    if (clonedData == NULL) {
        Message::error("l'allocation memoire pour l'image a echoue");
        return Image3D<Color>();
        }
    for (int c=0; c<w*h*d; c++)
        clonedData[c] = data[c];
    return Image3D<Color>(w, h, d, clonedData);
    }

template<class Color>
void Image3D<Color>::setBorderColor(Color c) {
    borderColor = c;
    }

template<class Color>
void Image3D<Color>::setWrapMode(WrapMode mode) {
    wrapMode = mode;
    }

template<class Color>
void Image3D<Color>::setupBorder(WrapMode mode, Color c) {
    wrapMode = mode;
    borderColor = c;
    }

template<class Color>
Color Image3D<Color>::sample(int i, int j, int k) const {
    switch (wrapMode) {
        case CLAMP_TO_BORDER :  return contains(i, j, k) ? texel(i,j,k) : borderColor;
        case REPEAT :           return texel(util::modulo(i, w), util::modulo(j, h), util::modulo(k, d));
        case MIRRORED_REPEAT :  return texel(util::mirror(i,w), util::mirror(j,h), util::mirror(k,d));
        case CLAMP_TO_EDGE :
        default :               return texel(util::clamp(i, 0, w-1), util::clamp(j, 0, h-1), util::clamp(k, 0, d-1));
        };
    }

template<class Color>
Color Image3D<Color>::interp(float x, float y, float z) const {
    int i0 = (int)(floorf(x - 0.5)),
        j0 = (int)(floorf(y - 0.5)),
        k0 = (int)(floorf(z - 0.5));
    float a = x - 0.5 - i0,
          b = y - 0.5 - j0,
          c = z - 0.5 - k0;
    return (1-a) * (1-b) * (1-c) * sample(i0, j0  , k0  ) + a * (1-b) * (1-c) * sample(i0+1, j0  , k0  )
         + (1-a) * (1-b) *    c  * sample(i0, j0  , k0+1) + a * (1-b) *    c  * sample(i0+1, j0  , k0+1)
         + (1-a) *    b  * (1-c) * sample(i0, j0+1, k0  ) + a *    b  * (1-c) * sample(i0+1, j0+1, k0  )
         + (1-a) *    b  *    c  * sample(i0, j0+1, k0+1) + a *    b  *    c  * sample(i0+1, j0+1, k0+1);
    }

/*********************************************************************************************/
// fonction de chargement de texture :

template<class Color>
void Image3D<Color>::loadTexture3D(GLint texFormat, GLenum target) const {
    glTexImage3D(target, 0, texFormat, w, h, d, 0, Color::DATA_FORMAT, Color::DATA_TYPE, data);
    }

template<class Color>
Image3D<Color> Image3D<Color>::readTexture(Texture *tex) {
    int w = tex->getWidth(),
        h = tex->getHeight(),
        d = tex->getDepth();
    Image3D<Color> res(w, h, d, NULL);
    if (w < 0 || h < 0 || d < 0)
        Message::error("dimensions de la texture non definies");
    else {
        tex->bind();
        glGetTexImage(GL_TEXTURE_3D, 0, Color::DATA_FORMAT, Color::DATA_TYPE, res.data);
        }
    return res;
    }

/*********************************************************************************************/
// fonctions de chargement/sauvegarde de fichiers :

template<class Color>
Image3D<Color>::Image3D(QString fileName) : AbstractImage3D(0,0,0), data(NULL), wrapMode(CLAMP_TO_EDGE) {
    QDir dir(fileName);
    if (!dir.exists()) {
        Message::error(QString("le sous-repertoire '%1' n'existe pas").arg(fileName));
        return;
        }
    dir.setNameFilters(QStringList("*.png"));
    QStringList fileNamesList = dir.entryList(QDir::Files, QDir::Name);
    
    d = fileNamesList.size();
    if (d == 0) {
        Message::error(QString("aucun fichier image dans le repertoire '%1'").arg(fileName));
        return;
        }
    
    QImage slice0(dir.filePath(fileNamesList[0]));
    w = slice0.width();
    h = slice0.height();
    data = new Color[w * h * d];
    if (data == NULL) {
        Message::error("memoire insuffisante pour charger l'image 3D en memoire");
        w = h = d = 0;
        return;
        }

    for (int k=0; k<d; k++) {
        QImage slice(dir.filePath(fileNamesList[k]));
        // on verifie que les slices ont bien la mme taille :
        if (slice.width() != w || slice.height() != h) {
            Message::error("les slices d'une image 3D doivent avoir toutes la mme taille");
            destroy();
            return;
            }
        // on recopie la slice dans l'image 3D :
        for (int i=0; i<w; i++) for (int j=0; j<h; j++)
            texel(i,j,k) = Color(slice.pixel(i, h-1-j));    // inversion des ordonnees
        }
    }

template<class Color>
void Image3D<Color>::save(QString dirName, QString subDirName) const {
    if (!loaded()) {
        Message::error("l'image ne contient pas de donnees");
        return;
        }
    
    QDir dir(dirName);
    if (!dir.exists()) {
        Message::error(QString("le repertoire '%1' n'existe pas").arg(dirName));
        return;
        }
    
    if (dir.exists(subDirName)) {
        // les eventuels fichiers que contient le sous-repertoire sont effaces :
        QDir subDir = dir;
        subDir.cd(subDirName);
        QStringList files = subDir.entryList(QDir::Files | QDir::NoSymLinks);
        foreach (QString file, files) subDir.remove(file);
        }
    else if (!dir.mkdir(subDirName)) {
        Message::error(QString("impossible de creer le repertoire '%1'").arg(subDirName));
        return;
        }
    
    QString absDirName = util::filePath(dirName, subDirName);
    
    for (int k=0; k<d; k++) {
        QImage slice(w, h, QImage::Format_ARGB32);
        for (int i=0; i<w; i++) for (int j=0; j<h; j++)
            slice.setPixel(i, h-1-j, texel(i,j,k).toQRgb());    // inversion des ordonnees
        QString fileName = util::filePath(subDirName, QString("slice%1.png").arg(k, 3, 10, QLatin1Char('0')));
        if (!image.save(fileName, "PNG"))
            Message::error(QString("probleme lors de la sauvegarde du fichier '%1'").arg(fileName));
        }
    }

}

