#ifndef __PROG_GLSL__
#define __PROG_GLSL__

#include "ShaderGLSL.h"
#include "TextureSet.h"
#include <QString>
#include <QStringList>
#include <QList>

namespace apig {
    
    class Uniform;
    class UniformGLSL;
    class ProgParam;
    
    /** This class represents a GLSL program.
     * A program is usually made of a vertex shader and a fragment shader.
     * Use static methods to load a program from GLSL source files.
     */
    class ProgGLSL {
        public:
            //ProgGLSL() : created(false), linked(false) {}
            //~ProgGLSL();
            //ProgGLSL();     // un contexte OpenGL doit etre actif
            static ProgGLSL* create();     // un contexte OpenGL doit etre actif, retourne NULL ssi une erreur est survenue
            ~ProgGLSL();
            
        // methodes de chargement :
        //-------------------------
            // les methodes suivantes retournent NULL si une erreur a survenu :
            static ProgGLSL* loadSubDir(QString rootName);                                  // ouvre une bote de dialogue et propose de choisir un sous-repertoire de <rootName>
            static ProgGLSL* loadDir(QString dirName);                                      // charge toutes les sources du sous-repertoire <dirName>
            static ProgGLSL* load(QString name);                                            // charge les fichiers <name>.vert et <name>.frag
            static ProgGLSL* load(QString vertName, QString fragName);                      // charge le vertex shader <vertName> et le fragment shader <fragName> (avec extensions)
            static ProgGLSL* loadFrag(QString dirName, QString fragName = QString::null);   // charge le fragment shader <dirName/fragName> avec le (seul) vertex shader se trouvant dans le mme repertoire
        
        // methodes pour l'activation du programme :
        //------------------------------------------
            void activate();    // active le programme et actualise les variables liees
            void deactivate();
            bool isLinked() const;          // true ssi le prog a ete linke correctement
            void activateTextures() const;  // doit tre appele pendant que le programme est active, si les textures attachees ont ete changees
            
        // methodes utiles :
        //------------------
            // la liste des sources de type <shaderType> presents dans le repertoire <dirName> :
            static QStringList sourcesList(QString dirName, ShaderGLSL::Type shaderType, bool relative=true);
    
        // methodes de verifications :
        //----------------------------
            void validate();            // valide et affiche des infos de validation
            void printInfo() const;     // affiche la liste d'attributs et uniforms actifs
            void printActiveAttribs() const;    // affiche la liste d'attributs actifs
            void printActiveUniforms() const;   // affiche la liste d'uniforms actifs
            
        // phases de chargement d'un programme (a appeler dans cet ordre) :
        //-----------------------------------------------------------------
            //void create();
            void bindShaderFile(QString fileName, ShaderGLSL::Type type);
            void bindShaderSource(QString source, ShaderGLSL::Type type);
            void link();
    
        // methodes d'actualisation de variables (a faire apres le linkage et avant la premiere activation) :
        //---------------------------------------------------------------------------------------------------
            // les variables seront automatiquement actualisees  l'activation du programme en utilisant la valeur pointee juste avant l'activation
            // si warn==true et la variable n'existe pas, un message de warning est affiche (utiliser warn=false pour les variables optionnelles)
            void addUniform(Uniform *uniform, bool warn=true, bool ownUniform=true);    // si ownUniform==true, la suppression de <uniform> est prise en charge
            void addUniform(QString name, const float  *value, bool warn=true);
            void addUniform(QString name, const int    *value, bool warn=true);
            void addUniform(QString name, const bool   *value, bool warn=true);
            void addUniformVec2(QString name, const float *value, bool warn=true);
            void addUniformVec3(QString name, const float *value, bool warn=true);
            void addUniformVec4(QString name, const float *value, bool warn=true);
            void addUniformIVec2(QString name, const int *value, bool warn=true);
            void addUniformIVec3(QString name, const int *value, bool warn=true);
            void addUniformIVec4(QString name, const int *value, bool warn=true);
            void addUniformBVec2(QString name, const bool *value, bool warn=true);
            void addUniformBVec3(QString name, const bool *value, bool warn=true);
            void addUniformBVec4(QString name, const bool *value, bool warn=true);
            void addUniformMat2(QString name, const float *value, bool warn=true);      // les coefficients de la matrice comme succession de colonnes (column-major order)
            void addUniformMat3(QString name, const float *value, bool warn=true);      // les coefficients de la matrice comme succession de colonnes (column-major order)
            void addUniformMat4(QString name, const float *value, bool warn=true);      // les coefficients de la matrice comme succession de colonnes (column-major order)
            void addTexture(QString name, Texture *texture, bool warn=true);
            
        // methodes pour changer la valeur d'une variable (a faire quand le programme est actif) :
        //----------------------------------------------------------------------------------------
            // si warn==true et la variable n'existe pas, un message de warning est affiche (utiliser warn=false pour les variables optionnelles)
            void setUniform(QString name, int   value, bool warn=true);
            void setUniform(QString name, float value, bool warn=true);
            void setUniform(QString name, bool value, bool warn=true);
            void setUniformVec2(QString name, float v0, float v1, bool warn=true);
            void setUniformVec3(QString name, float v0, float v1, float v2, bool warn=true);
            void setUniformVec4(QString name, float v0, float v1, float v2, float v3, bool warn=true);
            void setUniformVec2(QString name, const float *value, bool warn=true);
            void setUniformVec3(QString name, const float *value, bool warn=true);
            void setUniformVec4(QString name, const float *value, bool warn=true);
            void setUniformIVec2(QString name, int v0, int v1, bool warn=true);
            void setUniformIVec3(QString name, int v0, int v1, int v2, bool warn=true);
            void setUniformIVec4(QString name, int v0, int v1, int v2, int v3, bool warn=true);
            void setUniformIVec2(QString name, const int *value, bool warn=true);
            void setUniformIVec3(QString name, const int *value, bool warn=true);
            void setUniformIVec4(QString name, const int *value, bool warn=true);
            void setUniformBVec2(QString name, bool v0, bool v1, bool warn=true);
            void setUniformBVec3(QString name, bool v0, bool v1, bool v2, bool warn=true);
            void setUniformBVec4(QString name, bool v0, bool v1, bool v2, bool v3, bool warn=true);
            void setUniformBVec2(QString name, const bool *value, bool warn=true);
            void setUniformBVec3(QString name, const bool *value, bool warn=true);
            void setUniformBVec4(QString name, const bool *value, bool warn=true);
            void setUniformMat2(QString name, const float *value, bool warn=true);    // les coefficients de la matrice comme succession de colonnes (column-major order)
            void setUniformMat3(QString name, const float *value, bool warn=true);    // les coefficients de la matrice comme succession de colonnes (column-major order)
            void setUniformMat4(QString name, const float *value, bool warn=true);    // les coefficients de la matrice comme succession de colonnes (column-major order)
            
        private:
            ProgGLSL(GLuint id) : id(id), linked(false) {}
            void updateVariables() const;
            void clearParameters();
            UniformGLSL* createUniformGLSL(QString name, bool warn) const;          // retourne NULL si la variable <name> n'existe pas
            void bindShaderDir(QString dirName, ShaderGLSL::Type shaderType);       // <dirName> relatif au repertoire de base
            void attach(ShaderGLSL shader);
            void detach(ShaderGLSL shader);
            QString getInfoLog() const;
            GLint getProgParam(GLenum paramName) const;
            static QString typeEnumString(GLenum type);     // convertit en QString une enum representant un type GLSL
    
        private:
            GLuint id;                      // id OpenGL du programme
            bool linked;                    // true ssi le programme a ete linke avec succs
            QList<ShaderGLSL> shaders;      // liste des shaders formant ce programme
            QList<ProgParam*> parameters;   // la liste des variables attachees au programme
            TextureSet textures;            // la liste des textures attachees au programme
        };
    }

#endif

