#ifndef __VEC3__
#define __VEC3__

#include "Vec2.h"
#include <QString>
#include <QDomElement>
#include <QDomDocument>

namespace apig {
    class Vec4;

    /** Representation of a three-dimensional vector. */
    class Vec3 {
        public:
            Vec3(float x, float y, float z);
            Vec3(float c = 0);  // vecteur (c,c,c)
            Vec3(Vec2 v, float z = 0);
            Vec3(float x, Vec2 yz);
            Vec3(const float *v);   // tableau de 3 valeurs
            Vec3(Vec4 vec);         // fait la division par la 4eme composante avant d'affecter les composantes (x,y,z)
            Vec3(const QDomElement &element);
            
        // operateurs de conversion :
        //---------------------------
            // operateurs de conversion automatique :
            inline operator const float* () const { return v; }     // cast en const float*
            inline operator       float* ()       { return v; }     // cast en float*
            // autres conversions :
            QString toQString() const;
            Vec4    toVec4()    const;
            // lecture / ecriture d'un element XML :
            void initFromDOMElement(const QDomElement &element);
            QDomElement domElement(const QString &name, QDomDocument &document) const;
    
        // operations geometriques :
        //--------------------------
            float norm() const;         // norme
            float norm2() const;        // norme au carre
            void  normalize();          // normalise le vecteur
            Vec3  normalized() const;   // retourne le vecteur normalise
    
            friend Vec3  vec(const Vec3 &a, const Vec3 &b);    // produit vectoriel
            friend float dot(const Vec3 &a, const Vec3 &b);    // produit scalaire
            friend Vec3  operator^(const Vec3 &a, const Vec3 &b) { return vec(a,b); }
            friend float operator|(const Vec3 &a, const Vec3 &b) { return dot(a,b); }
            
            friend float dist(Vec3 v1, Vec3 v2);    // distance euclidienne
            friend Vec3 triangleNormal(const Vec3 &a, const Vec3 &b, const Vec3 &c);    // normale du triangle (a,b,c), dans le sens trigo
    
            Vec3 ortho() const;     // retourne un vecteur orthogonal a celui-ci
    
        // operateurs arithmetiques :
        //---------------------------
            // operateurs arithmetiques unaires :
            friend Vec3 operator-(const Vec3 &a);
    
            // operateurs arithmetiques binaires :
            friend Vec3 operator+(const Vec3 &a, const Vec3 &b);
            friend Vec3 operator-(const Vec3 &a, const Vec3 &b);
            friend Vec3 operator*(const Vec3 &a, const Vec3 &b);
            friend Vec3 operator/(const Vec3 &a, const Vec3 &b);
                
            // operations avec un scalaire :
            friend Vec3 operator*(float s, const Vec3 &a);
            friend Vec3 operator*(const Vec3 &a, float s);
            friend Vec3 operator/(float s, const Vec3 &a);
            friend Vec3 operator/(const Vec3 &a, float s);
            
            // operateurs de modification :
            Vec3& operator+=(const Vec3 &a);
            Vec3& operator-=(const Vec3 &a);
            Vec3& operator*=(const Vec3 &a);
            Vec3& operator/=(const Vec3 &a);
            Vec3& operator*=(float s);
            Vec3& operator/=(float s);
        
        // operateurs de comparaison :
        //----------------------------
            friend bool operator==(const Vec3 &a, const Vec3 &b);   // comparaison a 1e-10 pres
            friend bool operator!=(const Vec3 &a, const Vec3 &b);   // comparaison a 1e-10 pres
            // les comparaisons suivantes sont le <et> booleen des comparaisons sur les composantes :
            friend bool operator>=(const Vec3 &a, const Vec3 &b);
            friend bool operator<=(const Vec3 &a, const Vec3 &b);
            friend bool operator>(const Vec3 &a, const Vec3 &b);
            friend bool operator<(const Vec3 &a, const Vec3 &b);
            friend bool operator>=(const Vec3 &a, float b);
            friend bool operator<=(const Vec3 &a, float b);
            friend bool operator>(const Vec3 &a, float b);
            friend bool operator<(const Vec3 &a, float b);
        
        // autres fonctions pratiques :
        //-----------------------------
            friend Vec3 abs(const Vec3 &a);
            friend Vec3 sign(const Vec3 &a);
            friend Vec3 floor(const Vec3 &a);
            friend Vec3 ceil(const Vec3 &a);
            friend Vec3 fract(const Vec3 &a);
            friend Vec3 min(const Vec3 &a, const Vec3 &b);
            friend Vec3 min(const Vec3 &a, float b);
            friend Vec3 max(const Vec3 &a, const Vec3 &b);
            friend Vec3 max(const Vec3 &a, float b);
            friend Vec3 clamp(const Vec3 &a, const Vec3 &min, const Vec3 &max);
            friend Vec3 clamp(const Vec3 &a, float min, float max);
            friend Vec3 mix(const Vec3 &a, const Vec3 &b, const Vec3 &alpha);     // (1-alpha) * a + alpha * b
            friend Vec3 mix(const Vec3 &a, const Vec3 &b, float alpha);           // (1-alpha) * a + alpha * b
            friend Vec3 step(const Vec3 &e, const Vec3 &a);
            friend Vec3 step(float e, const Vec3 &a);
            friend float min(const Vec3 &a);    // minimum des composantes
            friend float max(const Vec3 &a);    // maximum des composantes
        
        // operateurs de swizzle :
        //------------------------
            Vec2 xx() const { return Vec2(x,x); }
            Vec2 xy() const { return Vec2(x,y); }
            Vec2 xz() const { return Vec2(x,z); }
            Vec2 yx() const { return Vec2(y,x); }
            Vec2 yy() const { return Vec2(y,y); }
            Vec2 yz() const { return Vec2(y,z); }
            Vec2 zx() const { return Vec2(z,x); }
            Vec2 zy() const { return Vec2(z,y); }
            Vec2 zz() const { return Vec2(z,z); }
        
            Vec3 yzx() const { return Vec3(y,z,x); }
            Vec3 zxy() const { return Vec3(z,x,y); }
            
        // calcul de vecteurs aleatoires (variables uniformes) :
        //------------------------------------------------------
            static Vec3 random();                       // valeurs dans [0,1[
            static Vec3 random(float min, float max);   // valeurs dans [min,max[
            static Vec3 random(Vec3 min, Vec3 max);     // bornes par composantes
            static Vec3 randomDir();                    // retourne un point aleatoire de la sphere unite
        
    
        // operations OpenGL :
        //--------------------
            void glVertex() const;    // effectue le glVertex correspondant
            void glNormal() const;    // effectue le glNormal correspondant
            
        public:
            union {
                struct { float x, y, z; };
                float v[3];
                };
        };
    
    }

#endif

